"""Definition of the CLI.

"""

import os
import argparse


def parse_args(description:str, args:iter=None) -> dict:
    return cli_parser(description).parse_args(args)

def existant_file(filepath:str) -> str:
    """Argparse type, raising an error if given file does not exists"""
    if not os.path.exists(filepath):
        raise argparse.ArgumentTypeError("file {} doesn't exists".format(filepath))
    return filepath

def writable_file(filepath:str) -> str:
    """Argparse type, raising an error if given file is not writable.
    Will delete the file !

    """
    try:
        with open(filepath, 'w') as fd:
            pass
        os.remove(filepath)
        return filepath
    except (PermissionError, IOError):
        raise argparse.ArgumentTypeError("file {} is not writable.".format(filepath))


def cli_parser(description:str) -> argparse.ArgumentParser:
    # main parser
    parser = argparse.ArgumentParser(description=description.strip())

    parser.add_argument('infile', type=existant_file,
                        help="Name of the input file to compress")
    parser.add_argument('--outfile', '-o', type=writable_file, default='out.bbl',
                        help="Name of the bubble file to produce")


    return parser



"""Test of graph compression.

Note that the test_* functions are generated
from the template_test_function function.

"""
import os
import pytest

from powergrasp.routines import compress_by_cc
from powergrasp import constants
from .test_cases import cases


def template_test_function(file:str, bubblelines:str):
    def test_function():
        found = unified_bubble(compress_by_cc(file))
        expected = unified_bubble(bubblelines.splitlines(keepends=False))
        assert found == expected
    return test_function


def unified_bubble(bubble_lines):
    """Return the set of comparable lines found in given bubble lines"""
    filtered = ['#']
    if not constants.BUBBLE_WITH_NODES:
        filtered.append('NODE')
    if not constants.BUBBLE_WITH_SETS:
        filtered.append('SET')
    filtered = tuple(filtered)

    # filter out comments and blank lines
    return set(line.strip() for line in bubble_lines
               if not line.startswith(filtered) and len(line.strip()) > 0)


for fname, bubble in cases.items():
    func = template_test_function('data/' + fname, bubble)
    globals()['test_' + os.path.basename(fname).replace('-', '_')] = func




from powergrasp.constants import TEST_INTEGRITY, SHOW_DEBUG, COVERED_EDGES_FROM_ASP


class Motif:
    """A motif found by a motif searcher.

    The motif is a proxy to the ASP model, with a large set of properties
    that Graph and Searchers will use to determine the Motif effect
    on the Graph.

    """
    def __init__(self, typename:str, atoms:dict, maximal:bool, step:int, searcher:object):
        self.typename, self.atoms, self.ismaximal, self.step = str(typename), dict(atoms), bool(maximal), int(step)
        self.step_modifier = 0
        self._searcher = searcher
        if SHOW_DEBUG:
            from pprint import pprint
            print('ATOMS FOR MOTIF {}:'.format(typename))
            pprint(self.atoms)
        self._score = None


    @property
    def name(self) -> str:  return self.typename
    @property
    def uid(self) -> int:  return self.step + self.step_modifier
    @property
    def type(self):  return self._searcher


    @property
    def score(self) -> int:
        if self._score is None:
            self._score = self._compute_score()
        return self._score
    @property
    def edge_cover(self) -> int:
        return self.score

    def _compute_score(self) -> int:
        """Compute and return the score of this motif"""
        score_atoms = tuple(self.atoms.get('score', ()))
        if not score_atoms:
            raise ASPModelError("No atom `score` found")
        if len(score_atoms) > 1:
            raise ASPModelError("Multiple atom `score` found: {}".format(len(score_atoms)))
        score_atom_args = score_atoms[0]
        if len(score_atom_args) != 1:
            raise ASPModelError("Atom `score` got a non-valid number of arguments: {}".format(score_atom_args))
        score = score_atom_args[0]
        if not isinstance(score, int):
            raise ASPModelError("Atom `score` got a non-int argument of type {}: {}".format(type(score), score))
        assert isinstance(score, int)
        return score

    def increase_step(self, step_diff:int):
        """Increase internally the step of self"""
        self.step_modifier += int(step_diff)

    @property
    def new_nodes(self) -> iter:
        yield from (node for numset, node in self.atoms.get('new_powernode', ()))
        yield from self.stars
    @property
    def new_powernodes(self) -> iter:
        yield from self.atoms.get('new_powernode', ())
    @property
    def powernodes(self) -> iter:
        _ = lambda s: s + (self.step_modifier if s == self.step else 0)
        yield from (
            (_(step), numset) for step, numset in self.atoms.get('powernode', ())
        )
    @property
    def stars(self) -> iter:
        yield from (args[0] for args in self.atoms.get('star', ()))
    @property
    def new_poweredge(self) -> iter:
        _ = lambda s: s + (self.step_modifier if s == self.step else 0)
        for args in self.atoms.get('poweredge', ()):
            if len(args) == 4:
                step_a, set_a, step_b, set_b = args
                yield (_(step_a), set_a), (_(step_b), set_b)
            elif len(args) == 3:
                step_a, set_a, node = args
                yield (_(step_a), set_a), node
    @property
    def hierachy_added(self) -> iter:
        _ = lambda s: s + (self.step_modifier if s == self.step else 0)
        yield from (
            (_(step1), numset, _(step2), numset2)
            for step1, numset, step2, numset2
            in self.atoms.get('hierarchy_add', ()))
    @property
    def hierachy_removed(self) -> iter:
        _ = lambda s: s + (self.step_modifier if s == self.step else 0)
        yield from (
            (_(step1), numset, _(step2), numset2)
            for step1, numset, step2, numset2
            in self.atoms.get('hierarchy_remove', ()))


    def edges_covered(self, sets:[frozenset]=None) -> iter:
        """If sets are given, the computation of edges covered by the motif
        is delegated to the searcher object.

        This allow to avoid a costly output from ASP.

        """
        if sets:  # give that to the searcher
            yield from frozenset(self.type.covered_edges(tuple(sets)))
        else:  # ASP provide us with the data
            yield from map(frozenset, self.atoms.get('covered_edge', ()))



"""PowerGrASP package.

"""

from . import cli
from .routines import compress_by_cc


def run_cli():
    args = cli.parse_args(__doc__)

    with open(args.outfile, 'w') as fd:
        for line in compress_by_cc(args.infile):
            fd.write(line + '\n')


if __name__ == "__main__":
    run_cli()



"""Test of ambiguous and multishot graph compression.

Note that the test_* functions are generated
from the template_test_function function.

"""
import os
import pytest
import itertools

from powergrasp.routines import compress_by_cc
from powergrasp import constants
from .ambiguous_test_cases import cases


def template_test_function(file:str, all_bubblelines:set):
    def test_function():
        found = unified_bubble(compress_by_cc(file))
        variants_keys = itertools.permutations(all_bubblelines['values'])
        variants = {all_bubblelines['common'] + '\n'
                    + all_bubblelines['variant'].format(**dict(zip('abcde', keys)))
                    for keys in variants_keys}
        expecteds = tuple(unified_bubble(variant.splitlines(keepends=False))
                          for variant in variants)
        # The following lines are here to get a proper understanding of the data
        #  in case of error.
        for expected in expecteds:
            if found == expected:
                return
        print('FOUND:', found)
        for idx, expected in enumerate(expecteds, start=1):
            print('EXPECTED', idx, ':', expected - found)
        assert found in expecteds, "Expected not found"
    return test_function


def unified_bubble(bubble_lines):
    """Return the set of comparable lines found in given bubble lines"""
    filtered = ['#']
    if not constants.BUBBLE_WITH_NODES:
        filtered.append('NODE')
    if not constants.BUBBLE_WITH_SETS:
        filtered.append('SET')
    filtered = tuple(filtered)

    # filter out comments and blank lines
    return frozenset(line.strip() for line in bubble_lines
                     if not line.startswith(filtered) and len(line.strip()) > 0)


for fname, bubbles in cases.items():
    func = template_test_function('data/' + fname, bubbles)
    globals()['test_' + os.path.basename(fname).replace('-', '_')] = func




from powergrasp.utils import quoted

def test_quoted():
    """Hard to test in docstring because of all the backslashs"""
    assert quoted('"a') == r'"\"a"'
    assert quoted('a"b') == r'"a\"b"'
    assert len(quoted('a"b')) == 6
    assert quoted(r'"\"a"') == r'"\"\"a\""'
    assert quoted(r'"\"a\""') == r'"\"\"a\"\""'
    assert len(quoted('"a')) == 5
    assert '{}'.format(quoted('"a')) == r'"\"a"'
    assert len('{}'.format(quoted('"a'))) == 5
    assert quoted(r'"\"a"') == r'"\"\"a\""'



"""Definition of high level functions operating the compression.

"""

from .searchers import CliqueSearcher, BicliqueSearcher
from .graph import Graph
from .constants import MULTISHOT_MOTIF_SEARCH


def compress(graph:Graph) -> [str]:
    """Yield bubble lines found in graph"""
    searchers = (CliqueSearcher(graph), BicliqueSearcher(graph))
    step = 0
    while True:
        step += 1
        score_to_beat = 0
        best_motif = None
        for searcher in sorted(searchers, key=lambda s: s.upperbound, reverse=True):
            motif = next(searcher.search(step, score_to_beat), None)
            if motif and (best_motif is None or motif.score > best_motif.score):
                best_motif, score_to_beat = motif, motif.score
        if best_motif:
            graph.compress(best_motif)
            for searcher in searchers:
                searcher.on_new_compressed_motif(best_motif)
        else:
            break  # nothing to compress
    yield from graph.bubble_repr()


def compress_multishot(graph:Graph) -> [str]:
    """Yield bubble lines found in graph"""
    from motif_batch import MotifBatch
    searchers = (CliqueSearcher(graph), BicliqueSearcher(graph))
    step = 0
    while True:
        step += 1
        score_to_beat = 0
        best_motifs, best_motifs_score = None, 0
        for searcher in sorted(searchers, key=lambda s: s.upperbound, reverse=True):
            motifs = MotifBatch(searcher.search(step, score_to_beat))
            if motifs and motifs.score > best_motifs_score:
                best_motifs, best_motifs_score = motifs, motifs.score
                score_to_beat = best_motifs_score
        if best_motifs:
            step += graph.compress_all(best_motifs.non_overlapping_subset())
            for searcher in searchers:
                searcher.on_new_compressed_motif(best_motifs)
        else:
            break  # nothing to compress
    yield from graph.bubble_repr()


def compress_by_cc(fname:str) -> [str]:
    """Yield bubble lines from compression of each cc found in given filename"""
    graphs = Graph.ccs_from_file(fname)
    for idx, graph in enumerate(graphs, start=1):
        if idx > 1:  yield ''
        yield '# CONNECTED COMPONENT {}'.format(idx)
        if MULTISHOT_MOTIF_SEARCH:
            yield from compress_multishot(graph)
        else:
            yield from compress(graph)




import re
import math
import phasme


def maximal_clique_size(nb_edge:int):
    """Maximal number of nodes implied in a clique when there is given
    number of edges between neighbors of a node.

    nb_edge -- number of edges between neighbors of a node n
    return -- maximal number of node implied in a clique implying node n

    The number of edges covered by a clique of n nodes is given by function k:
    k(n) = n(n-1) / 2

    The contraposition of k, giving the number of nodes n in a clique of k(n) edges is:
    n = (1 + √(1+8*k(n))) / 2

    Proof: function k may be expressed as:
    n^2 - n - 2k(n) = 0
    The discriminant Δ = (-1)^2 - 4 * 1 * (-2k(n)) = 1 + 8k(n)
    Roots are (1±√Δ)/2, only the positive one is relevant (unless dark matter is involved),
    therefore root of k is (1 + √(1+8*k(n))) / 2               □

    For any node n of a graph with v edges between the neighbors of n,
    the maximal clique MC implying node n is constitued of n
    and its neighbors having a link between them.
    An upperbound of that number of node is given by
    considering that neighbors link are optimally placed to create a clique.

    There is therefore, in the best case, NMC(n) nodes in the maximal
    clique implying n:
    NMC(n) = 1 + ⌊ (1 + √(1+8*k(n))) / 2 ⌋

    NMC(n) may then be used to compute the upperbound of clique search,
    by taking the maximal NMC(n) for all n of the graph in which a clique
    is searched.

    >>> maximal_clique_size(0)  # when no edge between neighbors: the node and one neighbor make the clique
    2
    >>> maximal_clique_size(2)  # when 2 edges between neighbors of n, the maximal clique imply two neighbors and n
    3
    >>> maximal_clique_size(3)
    4
    >>> maximal_clique_size(4)
    4
    >>> maximal_clique_size(5)
    4
    >>> maximal_clique_size(6)
    5

    """
    return 1 + int((1 + math.sqrt(1 + 8 * nb_edge)) / 2)


def quoted(string:str) -> str:
    """Return the given string, quoted, and with internal quotes escaped.

    Existing quotes will be escaped.

    >>> quoted('a')
    '"a"'
    >>> quoted('"a"')
    '"\\\\"a\\\\""'
    >>> quoted('"a')
    '"\\\\"a"'

    """
    # space is here to allow the regex to match for the first char
    return '"' + re.sub(r'([^\\])"', r'\1\\"', ' ' + string)[1:] + '"'


def unquoted(string:str) -> str:
    """Return the given string, unquoted.

    >>> unquoted('a')
    'a'
    >>> unquoted('"a"')
    'a'
    >>> unquoted('"a')
    '"a'

    """
    if string[0] == '"' and string[-1] == '"':
        return string[1:-1]
    return string


def normalized_name(string:str) -> str:
    """Return the normalized string, that consist of a quoted string
    with special characters replaced by their ord number.

    >>> normalized_name('a')
    '"a"'
    >>> normalized_name('A')
    '"A"'
    >>> normalized_name('"a"')
    '"a"'
    >>> normalized_name('"a')
    '"_c34_a"'

    """
    return quoted(phasme.commons.fixed_name(unquoted(string)))



"""List of compression cases.

"""

cases = {}  # filename: expected bubble

# Expected results of tested cases
cases['diacli.lp'] = """
NODE\tn
NODE\tr
NODE\tp
NODE\tq
NODE\te
NODE\tb
NODE\tm
NODE\ta
NODE\to
NODE\tc
NODE\tj
NODE\td
NODE\tg
NODE\tl
NODE\ts
NODE\tf
IN\tPWRN-a-3-1\tPWRN-a-1-1
IN\tPWRN-a-4-2\tPWRN-a-2-2
IN\ta\tPWRN-a-2-1
IN\tb\tPWRN-a-2-2
IN\tc\tPWRN-a-2-2
IN\td\tPWRN-a-2-1
IN\te\tPWRN-a-2-2
IN\tf\tPWRN-a-1-1
IN\tg\tPWRN-a-1-1
IN\tj\tPWRN-a-1-1
IN\tl\tPWRN-a-3-1
IN\tm\tPWRN-a-3-2
IN\tn\tPWRN-a-3-2
IN\to\tPWRN-a-4-2
IN\tp\tPWRN-a-3-1
IN\tq\tPWRN-a-4-2
IN\tr\tPWRN-a-1-1
IN\ts\tPWRN-a-3-1
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-2-1\t1.0
SET\tPWRN-a-2-2\t1.0
SET\tPWRN-a-3-1\t1.0
SET\tPWRN-a-3-2\t1.0
SET\tPWRN-a-3-2\t1.0
SET\tPWRN-a-4-2\t1.0
EDGE\tPWRN-a-3-1\tPWRN-a-3-2\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-1\t1.0
EDGE\tPWRN-a-3-2\tPWRN-a-4-2\t1.0
EDGE\tb\tc\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-2\t1.0
EDGE\tm\tn\t1.0
"""

cases['partition.lp'] = """
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
NODE\tf
NODE\tg
NODE\th
NODE\ti
IN\tc\tPWRN-a-2-1
IN\td\tPWRN-a-2-1
IN\tg\tPWRN-a-2-2
IN\th\tPWRN-a-2-2
IN\ta\tPWRN-a-1-1
IN\tb\tPWRN-a-1-1
IN\ti\tPWRN-a-1-1
IN\te\tPWRN-a-1-2
IN\tPWRN-a-2-2\tPWRN-a-1-2
IN\tf\tPWRN-a-1-2
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-1-2\t1.0
SET\tPWRN-a-2-1\t1.0
SET\tPWRN-a-2-2\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-2\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-2\t1.0
"""

cases['double_biclique_unambiguous.lp'] = """
NODE\ta
NODE\tb1
NODE\tb2
NODE\tc
NODE\td
NODE\td2
NODE\te
NODE\tf1
NODE\tf2
NODE\tg
NODE\th
NODE\ti
NODE\tj
NODE\tl
NODE\tm
NODE\tm2
NODE\tn
NODE\to
NODE\tp
NODE\tq
IN\tPWRN-a-3-1\tPWRN-a-2-1
IN\tPWRN-a-5-1\tPWRN-a-3-2
IN\ta\tPWRN-a-1-1
IN\tb1\tPWRN-a-1-2
IN\tb2\tPWRN-a-1-2
IN\tc\tPWRN-a-1-2
IN\td2\tPWRN-a-1-1
IN\td\tPWRN-a-1-1
IN\te\tPWRN-a-1-2
IN\tf1\tPWRN-a-2-1
IN\tf2\tPWRN-a-2-1
IN\tg\tPWRN-a-2-1
IN\th\tPWRN-a-2-2
IN\ti\tPWRN-a-2-2
IN\tj\tPWRN-a-2-1
IN\tl\tPWRN-a-3-1
IN\tm2\tPWRN-a-5-1
IN\tm\tPWRN-a-5-1
IN\tn\tPWRN-a-3-2
IN\to\tPWRN-a-1-2
IN\tp\tPWRN-a-3-1
IN\tq\tPWRN-a-1-2
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-1-2\t1.0
SET\tPWRN-a-2-1\t1.0
SET\tPWRN-a-2-2\t1.0
SET\tPWRN-a-3-1\t1.0
SET\tPWRN-a-3-2\t1.0
SET\tPWRN-a-5-1\t1.0
EDGE\tPWRN-a-5-1\tn\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-2\t1.0
EDGE\tf1\tg\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-2\t1.0
EDGE\tPWRN-a-3-1\tPWRN-a-3-2\t1.0
EDGE\tPWRN-a-3-2\tq\t1.0
EDGE\tb1\tc\t1.0
"""


cases['testblocks.lp'] = """
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
NODE\tf
NODE\tg
NODE\th
NODE\ti
NODE\tj
NODE\tl
NODE\tm
IN\te\tPWRN-a-2-1
IN\tPWRN-a-3-1\tPWRN-a-2-1
IN\ta\tPWRN-a-3-1
IN\tc\tPWRN-a-3-1
IN\tb\tPWRN-a-3-1
IN\td\tPWRN-a-3-1
IN\tg\tPWRN-a-4-1
IN\tf\tPWRN-a-4-1
IN\th\tPWRN-a-4-1
IN\ti\tPWRN-a-1-1
IN\tj\tPWRN-a-1-1
IN\tPWRN-a-4-1\tPWRN-a-1-1
IN\tm\tPWRN-a-1-1
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-2-1\t1.0
SET\tPWRN-a-3-1\t1.0
SET\tPWRN-a-4-1\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-1\t1.0
EDGE\tPWRN-a-3-1\tf\t1.0
EDGE\tPWRN-a-4-1\tl\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-1\t1.0
"""

cases['pnode-to-clique.lp'] = """
NODE\tb
NODE\td
NODE\ta
NODE\tf
NODE\te
NODE\tc
IN\te\tPWRN-a-1-2
IN\tc\tPWRN-a-1-2
IN\td\tPWRN-a-1-2
IN\ta\tPWRN-a-1-1
IN\tf\tPWRN-a-1-1
IN\tb\tPWRN-a-1-1
SET\tPWRN-a-1-2\t1.0
SET\tPWRN-a-1-1\t1.0
EDGE\tPWRN-a-1-2\tPWRN-a-1-2\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-2\t1.0
"""

cases['clique.lp'] = """
NODE\tc
NODE\tb
NODE\ta
NODE\td
IN\tc\tPWRN-a-1-1
IN\tb\tPWRN-a-1-1
IN\ta\tPWRN-a-1-1
IN\td\tPWRN-a-1-1
SET\tPWRN-a-1-1\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-1\t1.0
"""

cases['star.lp'] = """
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
IN\tc\tPWRN-a-1-2
IN\td\tPWRN-a-1-2
IN\tb\tPWRN-a-1-2
IN\te\tPWRN-a-1-2
SET\tPWRN-a-1-2\t1.0
EDGE\tPWRN-a-1-2\ta\t1.0
"""

cases['concomp.lp'] = """
NODE\t23
NODE\t42
EDGE\t23\t42\t1.0
NODE\t1
NODE\t2
NODE\t3
NODE\t4
NODE\t5
NODE\t6
IN\t1\tPWRN-1-1-1
IN\t2\tPWRN-1-1-2
IN\t3\tPWRN-1-1-2
IN\t4\tPWRN-1-1-2
IN\t5\tPWRN-1-1-2
IN\t6\tPWRN-1-1-1
SET\tPWRN-1-1-1\t1.0
SET\tPWRN-1-1-2\t1.0
EDGE\tPWRN-1-1-1\tPWRN-1-1-2\t1.0
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
IN\tb\tPWRN-a-1-2
IN\tc\tPWRN-a-1-2
IN\td\tPWRN-a-1-2
IN\te\tPWRN-a-1-2
SET\tPWRN-a-1-2\t1.0
EDGE\tPWRN-a-1-2\ta\t1.0
"""

cases['prio_deg.lp'] = """
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
NODE\tf
NODE\tg
NODE\th
NODE\ts
IN\tg\tPWRN-a-2-1
IN\tf\tPWRN-a-2-1
IN\th\tPWRN-a-2-1
IN\tc\tPWRN-a-1-1
IN\tb\tPWRN-a-1-1
IN\ta\tPWRN-a-1-1
IN\te\tPWRN-a-1-2
IN\td\tPWRN-a-1-2
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-1-2\t1.0
SET\tPWRN-a-2-1\t1.0
EDGE\tPWRN-a-2-1\ts\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-2\t1.0
EDGE\ta\ts\t1.0
"""

cases['perfectfit.lp'] = """
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
NODE\tf
NODE\tg
NODE\th
NODE\ti
NODE\tj
NODE\tl
IN\tPWRN-a-2-1\tPWRN-a-1-2
IN\tb\tPWRN-a-1-1
IN\th\tPWRN-a-1-2
IN\ti\tPWRN-a-1-2
IN\tPWRN-a-3-1\tPWRN-a-1-1
IN\ta\tPWRN-a-1-1
IN\tc\tPWRN-a-3-1
IN\td\tPWRN-a-3-1
IN\te\tPWRN-a-2-1
IN\tf\tPWRN-a-2-1
IN\tg\tPWRN-a-2-1
IN\tj\tPWRN-a-2-2
IN\tl\tPWRN-a-2-2
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-1-2\t1.0
SET\tPWRN-a-2-1\t1.0
SET\tPWRN-a-2-2\t1.0
SET\tPWRN-a-3-1\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-2\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-2\t1.0
EDGE\tPWRN-a-2-2\tPWRN-a-3-1\t1.0
"""

cases['variable-name.gml'] = """
NODE\tA
NODE\tB
NODE\tC
NODE\tD
NODE\tE
SET\tPWRN-A-1-2\t1.0
IN\tC\tPWRN-A-1-2
IN\tB\tPWRN-A-1-2
IN\tD\tPWRN-A-1-2
EDGE\tA\tPWRN-A-1-2\t1.0
EDGE\tB\tE\t1.0
"""

cases['test.gml'] = """
NODE\tb
NODE\tc
NODE\td
NODE\tf
NODE\tg
NODE\tl
NODE\tm
NODE\tp
NODE\ts
NODE\tv
NODE\tw
IN\td\tPWRN-b-1-1
IN\tw\tPWRN-b-1-1
IN\tc\tPWRN-b-1-1
IN\ts\tPWRN-b-1-1
IN\tb\tPWRN-b-1-1
IN\tf\tPWRN-b-2-2
IN\tPWRN-b-3-2\tPWRN-b-2-2
IN\tg\tPWRN-b-2-2
IN\tm\tPWRN-b-3-2
IN\tv\tPWRN-b-3-2
SET\tPWRN-b-1-1\t1.0
SET\tPWRN-b-2-2\t1.0
SET\tPWRN-b-3-2\t1.0
EDGE\tPWRN-b-3-2\tl\t1.0
EDGE\tl\tp\t1.0
EDGE\tPWRN-b-1-1\tPWRN-b-1-1\t1.0
EDGE\tm\tv\t1.0
EDGE\tc\tf\t1.0
EDGE\tPWRN-b-2-2\tb\t1.0
"""

cases['unclique.lp'] = """
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
NODE\tf
NODE\tg
NODE\th
NODE\ti
IN\tc\tPWRN-a-1-1
IN\te\tPWRN-a-1-1
IN\tf\tPWRN-a-1-1
IN\tg\tPWRN-a-1-1
IN\th\tPWRN-a-1-2
IN\ti\tPWRN-a-1-2
IN\td\tPWRN-a-2-1
IN\tb\tPWRN-a-2-1
IN\ta\tPWRN-a-2-1
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-1-2\t1.0
SET\tPWRN-a-2-1\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-2\t1.0
EDGE\tPWRN-a-2-1\tc\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-1\t1.0
"""

cases['order.lp'] = """
NODE\tb
NODE\t10
NODE\t4
NODE\t5
NODE\t6
NODE\ta
IN\t4\tPWRN-a-1-1
IN\ta\tPWRN-a-1-1
IN\t6\tPWRN-a-1-1
IN\t5\tPWRN-a-1-1
SET\tPWRN-a-1-1\t1.0
EDGE\ta\tb\t1.0
EDGE\t10\tPWRN-a-1-1\t1.0
"""

cases['quoting.lp'] = """
NODE\ta
NODE\tb
NODE\tc
IN\tc\tPWRN-a-1-2
IN\tb\tPWRN-a-1-2
SET\tPWRN-a-1-2\t1.0
EDGE\tPWRN-a-1-2\ta\t1.0
"""

cases['test.graphml'] = """
NODE\t1
NODE\t2
NODE\t3
NODE\t4
IN\t1\tPWRN-1-1-1
IN\t4\tPWRN-1-1-1
IN\t3\tPWRN-1-1-2
IN\t2\tPWRN-1-1-2
SET\tPWRN-1-1-1\t1.0
SET\tPWRN-1-1-2\t1.0
EDGE\tPWRN-1-1-1\tPWRN-1-1-2\t1.0
"""

cases['one_edge.lp'] = """
NODE\ta
NODE\tb
EDGE\ta\tb\t1.0
"""

cases['test.gml'] = """
NODE\tb
NODE\tc
NODE\td
NODE\tf
NODE\tg
NODE\tl
NODE\tm
NODE\tp
NODE\ts
NODE\tv
NODE\tw
IN\tc\tPWRN-b-1-1
IN\tb\tPWRN-b-1-1
IN\tw\tPWRN-b-1-1
IN\td\tPWRN-b-1-1
IN\ts\tPWRN-b-1-1
IN\tg\tPWRN-b-2-2
IN\tf\tPWRN-b-2-2
IN\tPWRN-b-3-2\tPWRN-b-2-2
IN\tv\tPWRN-b-3-2
IN\tm\tPWRN-b-3-2
SET\tPWRN-b-1-1\t1.0
SET\tPWRN-b-2-2\t1.0
SET\tPWRN-b-3-2\t1.0
EDGE\tPWRN-b-2-2\tb\t1.0
EDGE\tc\tf\t1.0
EDGE\tPWRN-b-1-1\tPWRN-b-1-1\t1.0
EDGE\tm\tv\t1.0
EDGE\tl\tp\t1.0
EDGE\tPWRN-b-3-2\tl\t1.0
"""

cases['horrible_data.lp'] = """
NODE\t_c92__c34_echo_c32_coucou_c92__c34_
NODE\t_c91_a_c93_
NODE\t_c39_echo_c32_coucou_c39_
SET\tPWRN-_c39_echo_c32_coucou_c39_-1-1\t1.0
IN\t_c92__c34_echo_c32_coucou_c92__c34_\tPWRN-_c39_echo_c32_coucou_c39_-1-1
IN\t_c39_echo_c32_coucou_c39_\tPWRN-_c39_echo_c32_coucou_c39_-1-1
EDGE\tPWRN-_c39_echo_c32_coucou_c39_-1-1\t_c91_a_c93_\t1.0
NODE\t_c91_a_c44_b_c93_
NODE\t_c36_PYTHONPATH
EDGE\t_c36_PYTHONPATH\t_c91_a_c44_b_c93_\t1.0
"""

cases['inclusions.lp'] = """
NODE\t1
NODE\t2
NODE\t3
NODE\t5
NODE\t6
NODE\t7
NODE\t8
NODE\t9
NODE\t10
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
NODE\tf
NODE\tg
NODE\th
NODE\ti
NODE\tj
NODE\tk
NODE\tl
NODE\tm
NODE\tn
NODE\to
NODE\tp
NODE\tq
NODE\tr
NODE\tv
NODE\tv2
NODE\tw
NODE\tw2
IN\tPWRN-a-6-2\tPWRN-a-2-1
IN\tPWRN-a-6-1\tPWRN-a-4-1
IN\tPWRN-a-5-1\tPWRN-a-6-1
IN\tPWRN-a-4-1\tPWRN-a-1-1
IN\tPWRN-a-4-2\tPWRN-a-3-1
IN\t10\tPWRN-a-2-1
IN\t1\tPWRN-a-3-1
IN\t2\tPWRN-a-3-1
IN\t3\tPWRN-a-3-1
IN\t5\tPWRN-a-2-1
IN\t6\tPWRN-a-2-1
IN\t7\tPWRN-a-2-1
IN\t8\tPWRN-a-2-1
IN\t9\tPWRN-a-2-1
IN\ta\tPWRN-a-1-1
IN\tb\tPWRN-a-1-1
IN\tc\tPWRN-a-1-1
IN\td\tPWRN-a-4-1
IN\te\tPWRN-a-6-1
IN\tf\tPWRN-a-5-1
IN\tg\tPWRN-a-5-1
IN\th\tPWRN-a-1-1
IN\ti\tPWRN-a-1-1
IN\tj\tPWRN-a-1-1
IN\tk\tPWRN-a-1-1
IN\tl\tPWRN-a-1-1
IN\tm\tPWRN-a-4-2
IN\tn\tPWRN-a-4-2
IN\to\tPWRN-a-4-2
IN\tp\tPWRN-a-4-2
IN\tq\tPWRN-a-6-2
IN\tr\tPWRN-a-6-2
IN\tv2\tPWRN-a-5-2
IN\tv\tPWRN-a-5-2
IN\tw2\tPWRN-a-5-2
IN\tw\tPWRN-a-5-2
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-2-1\t1.0
SET\tPWRN-a-3-1\t1.0
SET\tPWRN-a-4-1\t1.0
SET\tPWRN-a-5-1\t1.0
SET\tPWRN-a-6-1\t1.0
SET\tPWRN-a-4-2\t1.0
SET\tPWRN-a-5-2\t1.0
SET\tPWRN-a-6-2\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-1\t1.0
EDGE\tPWRN-a-2-1\tPWRN-a-2-1\t1.0
EDGE\tPWRN-a-3-1\tPWRN-a-3-1\t1.0
EDGE\tPWRN-a-4-1\tPWRN-a-4-2\t1.0
EDGE\tPWRN-a-5-1\tPWRN-a-5-2\t1.0
EDGE\tPWRN-a-6-1\tPWRN-a-6-2\t1.0
"""

cases['disjoint-subpnodes.lp'] = """
NODE\ta
NODE\tb
NODE\tc
NODE\td
NODE\te
NODE\tf
NODE\tg
NODE\th
IN\ta\tPWRN-a-1-1
IN\tb\tPWRN-a-3-1
IN\tc\tPWRN-a-3-1
IN\td\tPWRN-a-2-1
IN\te\tPWRN-a-2-1
IN\tf\tPWRN-a-2-1
IN\tPWRN-a-2-1\tPWRN-a-1-1
IN\tPWRN-a-3-1\tPWRN-a-1-1
SET\tPWRN-a-1-1\t1.0
SET\tPWRN-a-2-1\t1.0
SET\tPWRN-a-3-1\t1.0
EDGE\tPWRN-a-1-1\tPWRN-a-1-1\t1.0
EDGE\tPWRN-a-2-1\tg\t1.0
EDGE\tPWRN-a-3-1\th\t1.0
"""




import itertools
from collections import defaultdict

from . import asp
from . import utils
from .motif import Motif
from .graph import Graph
from .constants import TEST_INTEGRITY, SHOW_STORY, SHOW_DEBUG, MULTISHOT_MOTIF_SEARCH
from . import ASP_FILES


MOTIF_ASP_FILES = ASP_FILES['process-motif'], ASP_FILES['scoring_powergraph']
CLIQUE_ASP_FILES = (ASP_FILES['search-clique'], *MOTIF_ASP_FILES)
BICLIQUE_ASP_FILES = (ASP_FILES['search-biclique'], *MOTIF_ASP_FILES)


class MotifSearcher:
    """A motif searcher instance provides a search over a graph
    of a particular motif.
    If the graph must change more than described by signals,
    the motif searcher instance is invalid.

    It consists in the following client API:

    - init_for_graph: called on the graph on which the search will take place.
    - search: called to search a motif in the graph.
    - on_new_compressed_motif: called when a new motif is compressed. Information on that motif is given, notably its score.
    - covered_edges: function giving the edges covered by a given motif
    - name: the name of the motif, human-readable, making it different from the others.

    And the following methods to be overwritten (if no default) by subclasses:

    - _compute_initial_upperbound: give upperbound at first step for given graph (default: number of graph edges)
    - _compute_initial_lowerbound: idem for lowerbound (default: 2)
    - _search: called to search a motif in the graph, returns atoms found by ASP solver.

    Note that a motif searcher will maintain by itself the lowerbound
    and upperbound for the motifs, by considering that a motif score is the maximal it can do.
    Computations of these bounds may rely on complex computations
    and information taken from the graph receive by init.

    """

    def __init__(self, graph:Graph):
        self.graph = graph
        self.init_for_graph(graph)

    def init_for_graph(self, graph:Graph):
        self._upperbound = self.compute_initial_upperbound(graph)
        self._lowerbound = self.compute_initial_lowerbound(graph)
        if SHOW_STORY and self.lowerbound > self.upperbound:
            print("INFO Search for {} will not be functional, because bounds"
                  " ({};{}) avoid any search.".format(self.name, *self.bounds))

    @property
    def lowerbound(self) -> int:  return self._lowerbound
    @property
    def upperbound(self) -> int:  return self._upperbound
    @property
    def bounds(self) -> (int, int):  return self.lowerbound, self.upperbound

    def compute_initial_upperbound(self, graph:Graph) -> int:
        return graph.nb_edge
    def compute_initial_lowerbound(self, graph:Graph) -> int:
        return 2
    def on_new_compressed_motif(self, motif:Motif):
        """How to react when a motif is compressed ?"""
        if motif.ismaximal and motif.name == self.name:
            self._upperbound = motif.score
        self._lowerbound = 2  # now the optimization is not valid


    def search(self, step:int, score_to_beat:int=0) -> [Motif]:
        """Search for motifs, better than the one to beat."""
        lowerbound = max(self.lowerbound, score_to_beat)
        if lowerbound > self.upperbound:
            if SHOW_STORY:
                print("INFO No {} search because of bounds ({};{})."
                      "".format(self.name, *self.bounds))
            return  # impossible to find a motif in such conditions
        models = self._search(step, self.graph, lowerbound, self.upperbound)
        yield from (
            Motif(self.name, model, maximal=True, step=step, searcher=self)
            for model in models
        )

    def _search(self, graph:Graph, lowerbound:int, upperbound:int) -> Motif:
        raise NotImplementedError()


    def covered_edges(self, motif:Motif) -> iter:
        """Yield edges covered by given motif."""
        raise NotImplementedError()


    @property
    def name(self) -> str:
        return str(self._name())
    def _name(self) -> str:
        raise NotImplementedError()


class BicliqueSearcher(MotifSearcher):
    """Searcher for Bicliques."""

    def _name(self) -> str: return 'biclique'

    def compute_initial_lowerbound(self, graph:Graph) -> int:
        """Maximal lowerbound is the score of biggest star"""
        return max(len(neighbors) for _, neighbors in graph.neighbors())

    def _search(self, step:int, graph:Graph, lowerbound:int, upperbound:int) -> iter:
        graph = ''.join(graph.as_asp(step))
        if SHOW_DEBUG:
            print('UHJGMR: GRAPH:', graph)
        yield from asp.solve_motif_search(step, lowerbound, upperbound,
                                          files=BICLIQUE_ASP_FILES, graph=graph)

    def covered_edges(self, sets:[frozenset]) -> iter:
        """Return the edges that are covered by given sets"""
        assert len(sets) == 2
        yield from map(frozenset, itertools.product(*sets))


class CliqueSearcher(MotifSearcher):
    """Searcher for Cliques."""

    def _name(self) -> str: return 'clique'

    def compute_initial_upperbound(self, graph:Graph) -> int:
        """Use utils.maximal_clique_size function on all nodes of graph to
        compute the biggest clique possible theoretically."""
        clique_size = 0
        for node, _, nb_edge in graph.neighbors(increasing_degree=True, nb_edges_between_neighbors=True):
            node_clique_size = utils.maximal_clique_size(nb_edge)
            clique_size = max(clique_size, node_clique_size)
            # if clique_size > len(neighbors):  # TODO: test and prove useful that optimization
                # break  # we can't found better since it's sorted
        # return the number of edges, not the number of node
        return int(clique_size * (clique_size - 1) / 2)

    def _search(self, step:int, graph:Graph, lowerbound:int, upperbound:int) -> iter:
        graph = ''.join(graph.as_asp(step))
        yield from asp.solve_motif_search(step, lowerbound, upperbound,
                                          files=CLIQUE_ASP_FILES, graph=graph)

    def covered_edges(self, sets:[frozenset]) -> iter:
        """Return the edges that are covered by given sets"""
        assert len(sets) == 1
        nodes = frozenset(next(iter(sets)))
        yield from map(frozenset, itertools.combinations(nodes, r=2))



"""Batch of concurrent motif on a graph.

Offers an API useful for compression routine (access to score),
internal integrity test (equivalency of all contained motifs),
and non-overlapping subset selection.

"""

from itertools import islice
from powergrasp.motif import Motif
from powergrasp.constants import TEST_INTEGRITY


class MotifBatch:
    """Container of Motif"""

    def __init__(self, motifs:iter):
        self.motifs = tuple(motifs)
        self.score = self.motifs[0].score if self.motifs else None
        if TEST_INTEGRITY and self.score:
            scores = set(m.score for m in self.motifs)
            assert len(scores) == 1, "Multiple different scores in motifs: " + str(scores)

    def __bool__(self) -> bool:
        return bool(self.motifs)

    @property
    def empty(self) -> bool:
        return not self
    @property
    def name(self) -> str:  return self.motifs[0].typename
    @property
    def ismaximal(self) -> bool:  return self.motifs[0].ismaximal


    def non_overlapping_subset(self) -> iter:
        """Yield motifs contained in self that are non overlapping between them.

        These motifs can all be compressed at the same time because they
        do not share any node.

        """
        if self.empty: return  # nothing to yield
        motifs = iter(self.motifs)
        first = next(motifs)
        nodes = frozenset(first.new_nodes)
        yield first
        for motif in motifs:
            motif_nodes = frozenset(motif.new_nodes)
            if motif_nodes.isdisjoint(nodes):
                nodes |= motif_nodes
                yield motif



import pkg_resources
ASP_FILES = {
    fname: pkg_resources.resource_filename(__name__, 'asp/' + fname + '.lp')
    for fname in ('search-biclique', 'search-clique', 'process-motif', 'scoring_powergraph')
}

from .routines import compress_by_cc



import phasme
import networkx
from collections import defaultdict
from pprint import pprint
from powergrasp import utils
from powergrasp import constants
from powergrasp.motif import Motif
from powergrasp.constants import (TEST_INTEGRITY, SHOW_STORY, SHOW_DEBUG,
                                  SHOW_MOTIF_HANDLING, COVERED_EDGES_FROM_ASP)


def proper_nx_graph(graph:networkx.Graph) -> networkx.Graph:
    """Return the same graph, but with properly formatted node names."""
    def node_repr(node:str or int or float) -> str or int or float:
        """Ensure that nodes in graph are well encoded, ie that integer
        nodes are integer in graph, because ASP will understand
        and handle them as integer.
        """
        if isinstance(node, str) and node.isnumeric():
            return int(node)
        elif isinstance(node, int):
            return node
        # print('NODE REPR:', node, utils.normalized_name(node))
        return utils.normalized_name(node)  # needed to avoid collisions with constants, variable,…
    proper = type(graph)()
    repr_edges = (frozenset(map(node_repr, edge)) for edge in graph.edges)  # apply node representation
    proper.add_edges_from(edge for edge in repr_edges if len(edge) == 2)  # filter self loops
    return proper


class Graph:
    """A graph object, exposing some data on it.

    Note that a data is computed during its first access.

    Note the uid parameter for constructor

    """
    def __init__(self, graph:str or networkx.Graph):
        if isinstance(graph, networkx.Graph):
            nxgraph = networkx.Graph(graph)
        elif isinstance(graph, str):
            nxgraph = phasme.build_graph.graph_from_file(graph)
        else:
            raise ValueError("Unexpected {}".format(graph))
        # internal graph representation
        self.__edges = map(frozenset, nxgraph.edges)
        if TEST_INTEGRITY:
            self.__edges = tuple(self.__edges)
            for args in self.__edges:
                if len(args) > 2:
                    print('WARNING: Weird edge: {}. It will be filtered.'.format(args))
                else:
                    assert len(args) == 2, args
        self.__edges = set(edge for edge in self.__edges if len(edge) == 2)

        # data
        self.__nodes = set(nxgraph.nodes)
        self.__uid = str(min(self.__nodes, key=str))
        self.__nb_node = len(self.__nodes)
        self.__nb_cc = networkx.number_connected_components(nxgraph)
        self._nxgraph = networkx.freeze(nxgraph)

        self.__hierarchy = set()  # inclusions between powernodes
        self.__powernodes = defaultdict(set)  # (step, set) -> {node in powernode}
        self.__poweredges = defaultdict(set)  # (step, set) -> (step, set)

    @staticmethod
    def ccs_from_file(filename:str) -> iter:
        """Yield graphs found in given filename. Each graph is a connected component."""
        nxgraph = proper_nx_graph(phasme.build_graph.graph_from_file(filename))
        yield from (Graph(nxgraph.subgraph(cc))
                    for cc in networkx.connected_components(nxgraph))

    def as_asp(self, step:int, powerobjects:bool=False) -> [str]:
        """Yield atoms of graph's ASP string representation, in edges and hierarchy.

        powerobjects -- also include powernodes and poweredges

        """
        assert step > 0, step
        assert (step-1) >= 0, step
        for source, target in self.__edges:
            yield 'edge({},{}).'.format(source, target)
        for (stepa, seta), nodes in self.__powernodes.items():
            for node in nodes:
                yield 'block({},{},{},{}).'.format(step-1, stepa, seta, node)
        for stepa, seta, stepb, setb in self.__hierarchy:
            yield 'include_block({},{},{},{},{}).'.format(step-1, stepa, seta, stepb, setb)
        if powerobjects:
            raise NotImplementedError()


    def compress(self, motif:Motif):
        if SHOW_STORY or SHOW_DEBUG:
            print('COMPRESS…')
        if SHOW_DEBUG:
            pprint(dict(self.__powernodes))
        # Add the poweredges
        for source, target in motif.new_poweredge:
            if SHOW_MOTIF_HANDLING: print('\tP-EDGE', source, target)
            self.__poweredges[source].add(target)
        # Get the powernodes as set of nodes, and build the new ones.
        if not COVERED_EDGES_FROM_ASP:
            powernodes = set(motif.powernodes) | set(motif.stars)  # set of nodes
        for numset, node in motif.new_powernodes:
            uid = motif.uid, numset
            if SHOW_MOTIF_HANDLING: print('\tP-NODE', *uid, node)
            self.__powernodes[uid].add(node)
            if not COVERED_EDGES_FROM_ASP: powernodes.add(uid)

        # graph edges reduction and monitoring
        if COVERED_EDGES_FROM_ASP:
            covered = frozenset(motif.edges_covered())
        else:
            covered = frozenset(motif.edges_covered(self.__powernodes.get(uid, {uid}) for uid in powernodes))
        nb_edges = len(self.__edges)
        if TEST_INTEGRITY:
            edges = frozenset(self.__edges)
            if SHOW_DEBUG: print('IFBTVC GRAPH:', self.__edges)
            if SHOW_DEBUG: print('ZEDRBM COVER:', covered)
        self.__edges -= covered
        diff = nb_edges - len(covered) != len(self.__edges)
        if TEST_INTEGRITY and diff:
            diff_cov = covered - edges
            diff_edg = edges - covered
            raise ValueError("{} edges yielded by {} searcher were not in the graph: {}.\n\n"
                             "{} edges in the graph were not in the {} searcher: {}."
                             "".format(len(diff_cov), motif.name, ', '.join(map(str, diff_cov)),
                                       len(diff_edg), motif.name, ', '.join(map(str, diff_edg))))
        elif diff:
            raise ValueError("Edges yielded by {} searcher were not in the graph."
                             " Rerun with TEST_INTEGRITY to get the problem."
                             "".format(motif.name))
        if SHOW_MOTIF_HANDLING:
            print('\tCOVER', len(covered))

        # Now the big part: hierarchy. ASP send patch to apply on it.
        for args in motif.hierachy_added:
            if SHOW_MOTIF_HANDLING:
                print('\tADD HIERARCHY', args)
            self.__hierarchy.add(args)
            step_parent, num_parent, step_son, num_son = args
            # nodes in the parent block must be moved to the new block.
            # without that, the node would be at the same time in two powernodes.
            nodes = self.__powernodes[step_parent, num_parent] & self.__powernodes[step_son, num_son]
            if nodes:
                self.__powernodes[step_parent, num_parent] -= nodes
                self.__powernodes[step_son, num_son] |= nodes
                if SHOW_MOTIF_HANDLING:
                    print('\tMOVE NODES: {} FROM {} TO {}'.format(nodes, (step_parent, num_parent), (step_son, num_son)))

        for args in motif.hierachy_removed:
            if SHOW_MOTIF_HANDLING:
                print('\tDEL HIERARCHY', args)
            self.__hierarchy.remove(args)

        # Checking that the total number of nodes didn't change,
        #  i.e. there is no node placed in two powernodes,
        #  would be a great idea, but there is no way to do that since there is
        #  no way to know how many nodes are on ground (not in any powernode)
        #  without using the self.__powernodes attribute.
        # import itertools
        # nb_nodes_in_pnodes = sum(1 for _ in itertools.chain.from_iterable(self.__powernodes.values()))
        # nb_nodes_in_ground = ...
        # assert nb_nodes_in_pnodes == self.nb_node, (nb_nodes_in_pnodes, self.nb_node)
        if SHOW_DEBUG:
            print('PNODES:')
            pprint(dict(self.__powernodes))

        if TEST_INTEGRITY:
            inclusions = defaultdict(set)
            for stp, nsp, sts, nss in self.__hierarchy:
                inclusions[stp, nsp].add((sts, nss))
            # reversed:
            parents = defaultdict(set)
            for parent, sons in inclusions.items():
                for son in sons:
                    parents[son].add(parent)
            # detect son with multiple parents
            multiple_parents = {son for son, parents in parents.items() if len(parents) > 1}
            if multiple_parents:
                print('ERROR MULTIPLE PARENT:', len(multiple_parents))
                pprint(multiple_parents)
                exit(1)


    def compress_all(self, motifs:iter) -> int:
        """Compress all given motifs."""
        for step_diff, motif in enumerate(motifs, start=0):
            motif.increase_step(step_diff)
            self.compress(motif)
        return step_diff


    @property
    def uid(self) -> str:
        return self.__uid
    @property
    def nb_edge(self) -> int:
        return len(self.__edges)
    @property
    def nb_node(self) -> int:
        return self.__nb_node
    @property
    def nb_cc(self) -> int:
        return self.__nb_cc
    def neighbors(self, increasing_degree:bool=False,
                  nb_edges_between_neighbors:bool=False) -> iter:
        """Yield pairs (node, {neighbor}).

        increasing_degree -- yield the pairs sorted
        nb_edges_between_neighbors -- yield (node, {neighbors}, N), with N the
                                      number of edges in graph with given neighbors.

        """
        def yield_data():
            for node, neighbors in self._nxgraph.adjacency():
                neighbors = tuple(neighbors)
                if nb_edges_between_neighbors:
                    yield node, neighbors, self._nxgraph.subgraph(neighbors).size()
                else:
                    yield node, neighbors
        if increasing_degree:
            data = sorted(tuple(yield_data()), key=lambda x: len(x[1]))
        else:
            data = yield_data()
        data = tuple(data)
        yield from data


    def output(self, filename:str):
        """Write in given filename the bubble representation of the graph"""
        if SHOW_DEBUG:
            print('\n\n###########################\n')
            print('PNODES:')
            pprint(dict(self.__powernodes))
            print('HIERAC:')
            pprint(self.__hierarchy)
            print('PEDGES:')
            pprint(dict(self.__poweredges))
            print(' EDGES:')
            pprint(self.__edges)


        with open(filename, 'w') as fd:
            for line in self.bubble_repr():
                fd.write(line + '\n')


    def bubble_repr(self) -> iter:
        """Yield lines of bubble representation"""
        _format_name = lambda x: format_name(format_name(None, self.uid), x)
        if constants.BUBBLE_WITH_NODES:
            for node in self.__nodes:
                yield 'NODE\t{}'.format(_format_name(node))
        if constants.BUBBLE_WITH_SETS:
            for uid in self.__powernodes:
                yield 'SET\t{}\t1.0'.format(_format_name(uid))
        for stepa, seta, stepb, setb in self.__hierarchy:
            if stepa == 0 and seta == 0:  # base level
                continue  # ignore it
            yield 'IN\t{}\t{}'.format(_format_name((stepb, setb)), _format_name((stepa, seta)))  # contained first
        for (step, numset), nodes in self.__powernodes.items():
            for node in nodes:
                yield 'IN\t{}\t{}'.format(_format_name(node), _format_name((step, numset)))
        for source, targets in self.__poweredges.items():
            for target in targets:
                source, target = sorted(tuple(map(_format_name, (source, target))))
                yield 'EDGE\t{}\t{}\t{}'.format(source, target, constants.BUBBLE_POWEREDGE_FACTOR)
        for source, target in self.__edges:
            source, target = sorted(tuple(map(_format_name, (source, target))))
            yield 'EDGE\t{}\t{}\t{}'.format(source, target, constants.BUBBLE_EDGE_FACTOR)


def format_name(cc:str, name:str or tuple):
    """Return string representation of powernode of given step and set nb.

    This string representation is not ASP valid,
     but can be used as name in most file formats :

        PWRN-<cc>-<step>-<num_set>

    >>> format_name(1, 'cc')
    'cc'
    >>> format_name(1, '"A"')
    'A'
    >>> format_name(1, 'A')
    'A'
    >>> format_name(42, (1, 4))
    'PWRN-42-1-4'
    >>> format_name(23, ('cc', 4))
    'PWRN-23-cc-4'
    >>> format_name(23, ('"cc"', 4))
    'PWRN-23-cc-4'

    """
    if isinstance(name, int):
        return str(name)
    elif isinstance(name, str):
        if constants.BUBBLE_SIMPLIFY_QUOTES and name[0] == '"' and name[-1] == '"':
            return name[1:-1]
        else:
            return name
    return 'PWRN-{}-{}-{}'.format(cc, *map(lambda x: format_name(None, x), name))



"""Some constants to tune the compression behavior.

Mostly used for dev and debug.

"""
import os
import json

# Default values
constants = {
    # Run some integrity tests at some points. May slow the compression a lot.
    'TEST_INTEGRITY': True,

    # Show main steps of the compression.
    'SHOW_STORY': True,

    # Show motifs transformation into powergraph.
    'SHOW_MOTIF_HANDLING': False,

    # Show full trace of the compression. Useful for debugging.
    'SHOW_DEBUG': False,

    # Recover covered edges from ASP. If falsy, will ask motif searcher to compute the edges, which may be quicker.
    'COVERED_EDGES_FROM_ASP': False,

    # Nodes and sets are optional in the output bubble.
    'BUBBLE_WITH_NODES': True,
    'BUBBLE_WITH_SETS': True,

    # Edges in bubble are associated to a factor
    'BUBBLE_POWEREDGE_FACTOR': '1.0',
    'BUBBLE_EDGE_FACTOR': '1.0',

    # When possible, delete the quotes around identifiers in the bubble. May lead to node name collision.
    'BUBBLE_SIMPLIFY_QUOTES': True,

    # Change them according to config file
    'CONFIG_FILE': 'powergrasp.cfg',

    # Search for multiple motif in a single search. Accelerate the solving for graph with lots of equivalent motifs.
    'MULTISHOT_MOTIF_SEARCH': False,
}


try:
    with open(constants['CONFIG_FILE']) as fd:
        cfg = json.load(fd)
        for field, value in cfg.items():
            field = field.upper().replace(' ', '_')
            if field in constants:
                constants[field] = value
        print('INFO: config file {} loaded.'.format(fd.name))
except FileNotFoundError:
    if constants['SHOW_STORY'] or constants['SHOW_DEBUG']:
        print("INFO no config file")


# Put them in global access
for field, value in constants.items():
    globals()[field] = value



"""Search for motif using clyngor/clingo/ASP.

Function solve_motif_search is defined according to global constants.

"""

import math
import clyngor
from powergrasp.constants import COVERED_EDGES_FROM_ASP, SHOW_STORY, SHOW_DEBUG, MULTISHOT_MOTIF_SEARCH


def _build_solver(step:int, lowerbound:int, upperbound:int, files:iter, graph:str, options:str='') -> iter:
    """Return iterator over found models"""
    constants = {'k': step, 'lowerbound': lowerbound, 'upperbound': upperbound}
    if COVERED_EDGES_FROM_ASP:
        constants['covered_edges_from_asp'] = 1
    models = clyngor.solve(files=tuple(files), inline=str(graph), constants=constants, stats=False, options=options)
    if SHOW_STORY:
        print('SOLVE', models.command)
    return models.by_predicate.careful_parsing


def oneshot_motif_search(step:int, lowerbound:int, upperbound:int, files:iter, graph:str) -> iter:
    """Return iterable over the generator of the one best model
    containing atoms found in best model"""
    model = None
    for model in _build_solver(step, lowerbound, upperbound, files, graph):
        pass  # get the last one
    if model:
        yield model


def multishot_motif_search(step:int, lowerbound:int, upperbound:int, files:iter, graph:str) -> iter:
    """Yield atoms found in bests models"""
    all_models = _build_solver(step, lowerbound, upperbound, files, graph, options='--opt-mode=optN')
    best_opt, models = math.inf, []
    for model, opt in all_models.with_optimization:
        if SHOW_DEBUG:
            print('OPT, MODEL:', opt[0], model)
        if opt[0] < best_opt:  # smaller is best
            best_opt, models = opt[0], []  # model will be given again as last model, so no need to include it twice
        else:
            models.append(model)
    yield from models


# define the default behavior
if MULTISHOT_MOTIF_SEARCH:
    solve_motif_search = multishot_motif_search
else:
    solve_motif_search = oneshot_motif_search
